from base64 import b64encode, b64decode

from .utils import Singleton, safe_eval, FallbackBox, on_off
from .validator import Validator


class PayloadSelector(Singleton):
    """
    Examples:

        >>> dut = PayloadSelector.instance
        >>> dut.eval_selector('payload.k1', dict(k1='k1', k2='k2'))
        'k1'
        >>> dut.eval_selector('suppress_me if int(payload) == 0 else int(payload)', 1)
        1
        >>> dut.eval_selector('suppress_me if int(payload) == 0 else int(payload)', 0) is dut.SuppressLiteral
        True
        >>> dut.eval_selector('on_off(data)', True)
        'on'

        >>> dut.instance.register_custom_global('on_off_custom', lambda x: 'on' if x else 'off')
        >>> dut.eval_selector('on_off_custom(data)', True)
        'on'
        >>> dut.eval_selector('on_off_custom(data)', False)
        'off'
    """

    def __init__(self):
        self._suppress_literal = object()
        self._custom = {}
        self._register_globals()

    @property
    def SuppressLiteral(self):
        return self._suppress_literal

    @property
    def suppress_aliases(self):
        uppers = ["SUPPRESS", "SUPPRESSME", "SUPPRESSPUSH", "SUPPRESS_ME", "SUPPRESS_PUSH"]
        return uppers + [item.lower() for item in uppers]

    def _register_globals(self):
        self._custom["b64encode"] = b64encode
        self._custom["b64decode"] = b64decode
        self._custom["on_off"] = on_off

    def register_custom_global(self, name, fun):
        """
        Register a custom function that should be available in the context of the selector.

        Args:
            name: Name of the function to register.
            fun: The actual function to register.

        Returns:
            None.
        """
        Validator.is_function(fun=fun)
        Validator.is_instance(str, name=name)
        if name not in self._custom:  # Overriding a already existing custom will silently fail!
            self._custom[name] = fun

    def eval_selector(self, selector, payload):
        # Wrap payload inside a Box -> this makes dot accessable dictionaries possible
        # We create a dictionary cause payload might not be an actual dictionary.
        # This way wrapping with Box will always work ;-)
        payload = FallbackBox({'base': payload}).base
        if selector is None:
            return payload

        suppress_kwargs = {alias: self.SuppressLiteral for alias in self.suppress_aliases}
        return safe_eval(
            source=selector,
            payload=payload,
            data=payload,
            **suppress_kwargs,
            **self._custom
        )



from collections import namedtuple
from typing import Dict, List, Optional

from box import Box

from .plugins import load_plugin
from .plugins.pull import PullBase
from .plugins.push import PushBase

Task = namedtuple("Task", ["name", "pull", "pushes"])
Pull = namedtuple("Pull", ["instance"])
Push = namedtuple("Push", ["instance", "selector", "unwrap", "deps"])

TaskName = str
TaskCfg = Box
TaskSet = Dict[str, Task]


def _mk_pull(task: TaskCfg, **extra) -> Pull:
    args = {'name': '{task.name}_pull'.format(task=task), **extra, **task.pull.args}
    return Pull(instance=load_plugin(task.pull.plugin, PullBase, **args))


def _mk_push(task: TaskCfg, **extra) -> List[Push]:
    def _many(pushlist, prefix):
        for i, push in enumerate(pushlist):
            push_name = '{prefix}_{i}'.format(**locals())
            args = {'name': push_name, **extra, **push.args}
            unwrap = getattr(push, 'unwrap', False)
            yield Push(
                instance=load_plugin(push.plugin, PushBase, **args),
                selector=push.selector,
                unwrap=unwrap,
                deps=list(_many(push.deps, push_name))
            )
    return list(_many(task.pushes, "{task.name}_push".format(**locals())))


def from_dict(task: TaskCfg, base_path: Optional[str] = None) -> Task:
    if not isinstance(task, Box):
        task = Box(dict(base=task)).base

    extra = {'base_path': base_path}
    return Task(
        name=task.name,
        pull=_mk_pull(task, **extra),
        pushes=list(_mk_push(task, **extra))
    )


setattr(Task, 'from_dict', from_dict)


def tasks_to_str(tasks: TaskSet):
    res = []
    for _, t in tasks.items():
        res.append('{task.name} {{'.format(task=t))
        res.append('\tpull = {task.pull}'.format(task=t))
        res.append('\tpushes = ['.format())
        for push in t.pushes:
            res.append('\t\t{push}'.format(push=push))
        res.append('\t]')
        res.append('}')
    return "\n".join(res)
